"""d810.ctree: Ctree (decompiler AST) pattern matching and manipulation.

This package provides a pattern DSL for matching and transforming
Hex-Rays ctree nodes, ported from herast.

Sub-packages:
    patterns    - Pattern classes for matching ctree items
"""
