"""Pattern matching optimizer handlers.

Rules in this module are automatically discovered and registered by the
reloadable scanner. Both old (ast-based) and refactored (DSL-based) rules
are registered automatically when their modules are imported.
"""
