"""Cython-optimized performance modules for d810.

This package contains Cython implementations that mirror the d810 package
structure. These modules are used when Cython extensions are built and available.

If speedups are not installed, d810 automatically falls back to pure Python
implementations.
"""

__version__ = "0.1.0"
